/* ----------------------------------------------------------------------------
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 * --------------------------------------------------------------------------*/
#include <stdio.h>

/* sqlrext.h is required in all files using the RDM ODBC APIs */
#include "sqlrext.h"

/*! \file perf02.c
 * \brief Source code for the RDM Perf02 ODBC example
 */
 
/*!
 * \mainpage Perf02_ODBC
 * \author Daigoro F. Toyama, Raima Sr. Engineer
 * \date 2012
 *
 * \section Description
 * EXAMPLE - Perf02 ODBC
 *
 * The Perf02 example performs a number of atomic database operations
 * on a very simple database schema using the RDM ODBC APIs.
 *
 * Each time you run this example
 *   * The database is initialized (removing all existing data)
 *   * 50,000 new records are created
 *   * Each record is read
 *   * Each record is updated
 *   * Each record is read again
 *   * Each record is deleted
 *   * Database is cleaned up
 *
 * \section Perf02 Database
 * The example utilizes the SQL DDL syntax and specifies an inmemory database
 * with a single table (simple).  The table has a single column (int_col).
 * There are no index definitions for this database.
 * The SQL DDL commands for creating the database are executed in the 
 * initialization function, but the commands used to create the database can
 * also be found in the file perf02.sdl.
 *
 * \include perf02.sdl
 *
 * \section RDM ODBC Functions Used
 * The following RDM ODBC functions are illustrated:
 *   * SQLAllocHandle/SQLFreeHandle
 *   * SQLGetDiagRec
 *   * SQLConnect/SQLDisconnect
 *   * SQLExecDirect
 *   * SQLEndTran
 *   * SQLPrepare
 *   * SQLBindParameter
 *   * SQLExecute
 *   * SQLBindCol
 *   * SQLFetch
 *   * SQLEndTran
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 */

/*! \brief Create a test database
 *
 */
static SQLRETURN create_database(
    SQLHDBC hDbc)
{
    SQLRETURN sqlError;
    SQLHSTMT  hStmt;
    int16_t   ii;

    const char *cmds[] = {
        "CREATE DATABASE perf02_db",
        "CREATE TABLE simple (int_col INTEGER)",
        "COMMIT"
    };

    (void) SQLAllocHandle(SQL_HANDLE_STMT, hDbc, &hStmt);

    /* Attempt to drop the database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. */
    (void) SQLExecDirect(hStmt, (SQLCHAR *) "DROP DATABASE perf02_db", SQL_NTS);
    (void) SQLEndTran(SQL_HANDLE_DBC, hDbc, SQL_COMMIT);

    for (ii = 0; ii < sizeof(cmds) / sizeof(cmds[0]); ii++)
    {
        sqlError = SQLExecDirect(hStmt, (SQLCHAR *) cmds[ii], SQL_NTS);
        if (!SQL_SUCCEEDED(sqlError))
            break;
    }

    (void) SQLFreeHandle(SQL_HANDLE_STMT, hStmt);
    return SQL_SUCCESS;
}

/*! \brief Display a message and the elapsed time to perform database operations
 *
 */
void display_elapsed_time(
    const char *msg,         /**< [in] Message to be display along with the elapsed time */
    uint64_t    start_time,  /**< [in] Start of elapsed time to display */
    uint64_t    end_time)    /**< [in] End of elapsed */
{
    uint64_t elapsed_time = end_time - start_time;
    printf("%s" UINT64_SPEC " millisecs\n", msg, elapsed_time/1000000);
}

/*! \brief Display a RDM ODBC Error message
 *
 * This function retrieves error information from RDM ODBC and
 * displays it and the passed in message to stderr.
 */
void display_error_msg(
    SQLSMALLINT handle_type,  /**< [in] Type of ODBC Handle that triggered the error */
    SQLHANDLE   handle,       /**< [in] ODBC Handle that triggered the error */
    const char *msg           /**< [in] Message to be display along with ODBC error */
    )
{
    SQLCHAR SQLState[6];
    SQLINTEGER NativeError;
    SQLCHAR MessageText[100];
    SQLSMALLINT Len;

    SQLGetDiagRec(handle_type, handle, 1, SQLState, &NativeError, MessageText, sizeof(MessageText), &Len);
    
    fprintf(stderr, "%s: %s - %s (%d)\n", msg, SQLState, MessageText, NativeError);
}

/*! \brief Initialize the RDM ODBC library for use in the perf02 example
 *
 * This function 
 *   * Initializes the RDM Transactional File Server (TFS) to use the 
 *     single-user non-transactional STANDALONE implementation
 *   * Creates and intializes the simple database
 *   * Creates and connection an ODBC connection handle
 *   * Opens the "perf02" database in exclusive ('x') mode.
 *
 * @return Returns a SQLRETURN value (SQL_SUCCESS if successful)
 */
static SQLRETURN initialize(
    SQLHANDLE    *hENV,    /**< [out] Pointer to the environment handle to be initialized */
    SQLHANDLE    *hDBC)    /**< [out] Pointer to the connection handle to be initialized */
{
    SQLRETURN    sqlError;
    uint64_t     start_time;
    uint64_t     end_time;
    SQLHANDLE    hSTMT;
    DBCAT_STRUCT dbCat;

    *hENV = NULL;
    *hDBC = NULL;
    memset(&dbCat, 0, sizeof(DBCAT_STRUCT));

    start_time = psp_timeNanoSecs();

    /* Allocate the environment handle */
    sqlError = SQLAllocHandle(SQL_HANDLE_ENV, NULL, hENV);
    if (!SQL_SUCCEEDED(sqlError))
    {
        fprintf(stderr, "Error allocating an environment handle\n");
        return sqlError;
    }

    /* Allocate the connection handle */
    sqlError = SQLAllocHandle(SQL_HANDLE_DBC, *hENV, hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {       
        display_error_msg(SQL_HANDLE_ENV, *hENV, "Error allocating a connection handle");
        return sqlError;
    }

    /* Set the TFS type to STANDALONE */
    sqlError = SQLSetConnectAttr(*hDBC, SQL_ATTR_RDM_TFSINIT_TYPE,
            (SQLPOINTER) SQL_TFSTYPE_TFS, 0);
    if (!SQL_SUCCEEDED(sqlError))
    {       
        display_error_msg(SQL_HANDLE_DBC, *hDBC, "Error setting the TFS type");
        return sqlError;
    }

    /* Set the ACCESS_MODE (open mode) option in order to open the perf02_db
       database in exclusive mode */
    sqlError = SQLSetConnectAttr(*hDBC, SQL_ATTR_ACCESS_MODE,
            (SQLPOINTER) SQL_MODE_EXCLUSIVE, SQL_IS_UINTEGER);
    if (!SQL_SUCCEEDED(sqlError))
    {       
        display_error_msg(SQL_HANDLE_DBC, *hDBC, "Error setting the open mode");
        return sqlError;
    }

    /* Connect to the RDM SQL Server */
    sqlError = SQLConnect(*hDBC,
                            (SQLCHAR *) "", SQL_NTS,
                            (SQLCHAR *) "", SQL_NTS,
                            (SQLCHAR *) "", SQL_NTS);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_DBC, *hDBC, "Error calling SQLConnect");
        return sqlError;
    }

    /* Create the "perf02_db" database */
    sqlError = create_database(*hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_DBC, *hDBC, "Error creating a database");
        return sqlError;
    }

    /* Allocate a statement handle */
    sqlError = SQLAllocHandle(SQL_HANDLE_STMT, *hDBC, &hSTMT);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_DBC, *hDBC, "Error allocating a statement handle");
        return sqlError;
    }

    /* Initialize the database (creates new database files truncating any existing data) */    
    sqlError = SQLExecDirect(hSTMT, (SQLCHAR *) "INITIALIZE DATABASE perf02_db", SQL_NTS);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error initializing the perf02_db database");
        SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
        return sqlError;
    }

    (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);

    end_time = psp_timeNanoSecs();  
    
    /* Display information */
    printf("Initializing with STANDALONE configuration\n");
    display_elapsed_time("Preparing a new database:\t\t\t", start_time, end_time);
    printf("\tOne, simple record type.\n");
    printf("\tOn disk\n");
    printf("\tOpened exclusively\n");

    return sqlError;
}

/*! \brief Cleanup the RDM runtime library
 *
 * This functions closes all open databases and cleans up the RDM connection
 * handle used in the perf02 example.  It also terminates the STANDALONE TFS
 * used in the example. 
 */
static void cleanup(
    SQLHANDLE     hENV,    /**< [in] Environment handle to be cleaned up */
    SQLHANDLE     hDBC)    /**< [in] Connection handle to be cleaned up */
{
    uint64_t     start_time;
    uint64_t     end_time;

    start_time = psp_timeNanoSecs();
    if (hDBC)
    {
        /* Close any open databases */
        (void) SQLDisconnect(hDBC);

        /* Free the connection handle */
        (void) SQLFreeHandle(SQL_HANDLE_DBC, hDBC);
    }

    /* Free the environment handle */
    if (hENV)
        (void) SQLFreeHandle(SQL_HANDLE_ENV, hENV);

    end_time = psp_timeNanoSecs();

    display_elapsed_time("Cleanup:\t\t\t\t\t", start_time, end_time);
}

/*! \brief Insert 50,000 simple records into the perf02 database
 *
 * This function adds 50,000 simple records into the inmemory perf database.  The
 * simple table contains only a single integer column with no index.  The function will
 * calculate and display the total time elapsed while inserting the records
 *
 * @return Returns a SQLRETURN value (SQL_SUCCESS if successful)
 */
static SQLRETURN add_records(
    SQLHANDLE hDBC)         /**< [in] RDM ODBC Connection handle */
{
    int32_t   ii;
    uint64_t  start_time;
    uint64_t  end_time;
    SQLHANDLE hSTMT;
    SQLRETURN sqlError;
    int32_t   int_col = 0;

    start_time = psp_timeNanoSecs();

    /* allocates insert statement */
    sqlError = SQLAllocHandle(SQL_HANDLE_STMT, hDBC, &hSTMT);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_DBC, hDBC, "Error has occured allocating insert statement");
        return sqlError;
    }

    /* Prepare insert statement */
    sqlError = SQLPrepare(hSTMT, (SQLCHAR *) "INSERT INTO simple VALUES ?", SQL_NTS);
    if (!SQL_SUCCEEDED(sqlError))
    {        
        display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error has occured preparing the insert statement");
        SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
        return sqlError;
    }

    /* Bind the insert value to variable location */
    sqlError = SQLBindParameter(hSTMT, 1, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER, 0, 0, &int_col, 0, NULL);
    if (!SQL_SUCCEEDED(sqlError))
    {        
        display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error has occurred binding the insert statement parameter");
        SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
        return sqlError;
    }

    /* Loop to add 50,000 new simple record */
    for (ii = 0; ii < 50000; ii++)
    {
        /* Set values for the new record */
        int_col = 1;

        /* Add the record to the database */
        sqlError = SQLExecute(hSTMT);
        if (!SQL_SUCCEEDED(sqlError))
        {            
            display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error has occurred exeucting the insert statement");
            SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
            return sqlError;
        }
    }

    (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
    end_time = psp_timeNanoSecs();

    display_elapsed_time("Add 50,000 records:\t\t\t\t", start_time, end_time);
    return sqlError;
}

/*! \brief Read all simple records and calculate the sum of the int_col column
 *
 * This function reads all simple records stored in the perf02 database.  The
 * function will calculate the sum of the int_col field from all simple rows
 * in the database.  Note that it actually would be more efficient to use the
 * build in SUM aggregate function than to manually fetch each row and 
 * calculate the sum in the function ("SELECT SUM(int_col) from simple").  The
 * function will also calculate and display the total time elapsed while 
 * reading the records.
 *
 * @return Returns a SQLRETURN value (SQL_SUCCESS if successful)
 */
static SQLRETURN read_records(
    SQLHANDLE hDBC)         /**< [in] RDM ODBC Connection Handle */
{
    uint64_t          total = 0;
    uint64_t          start_time;
    uint64_t          end_time;
    SQLHANDLE         hSTMT;
    SQLRETURN         sqlError;
    int32_t           col_value;

    start_time = psp_timeNanoSecs();

    /* allocates select statement */
    sqlError = SQLAllocHandle(SQL_HANDLE_STMT, hDBC, &hSTMT);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_DBC, hDBC, "Error has occured allocating select statement");
        return sqlError;
    }

    /* Prepare and Execute the select statement */
    sqlError = SQLExecDirect(hSTMT, (SQLCHAR *) "select int_col from simple", SQL_NTS);
    if (!SQL_SUCCEEDED(sqlError))
    {        
        display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error has occured preparing/executing the select statement");
        (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
        return sqlError;
    }

    /* Bind the result column */
    sqlError = SQLBindCol(hSTMT, 1, SQL_C_SLONG, &col_value, sizeof(col_value), NULL);
    if (!SQL_SUCCEEDED(sqlError))
    {        
        display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error has occurred binding the select result column");
        (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
        return sqlError;
    }

    /* Loop through each simple record */
    while ((sqlError = SQLFetch(hSTMT)) == SQL_SUCCESS)
    {
        /* Add the value read to the running sum */
        total += col_value;
    }        

    /* We should break out of the loop when we get SQL_NO_DATA */
    if (sqlError == SQL_NO_DATA)
        sqlError = SQL_SUCCESS;
    else
    {
        display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error fetching rows from the simple table");
        (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
        return sqlError;
    }

    (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
    end_time = psp_timeNanoSecs();
    
    display_elapsed_time("Scanning, reading, summing 50,000 records:\t", start_time, end_time);
    printf("\tSum = " UINT64_SPEC "\n", total);

    return sqlError;
}

/*! \brief Update all simple records in the database
 *
 * This function iterates updates all simple records stored in the perf02
 * database.  The function will calculate and display the total time elapsed
 * while deleting the records. 
 *
 * @return Returns a SQLRETURN value (SQL_SUCCESS if successful)
 */
static SQLRETURN update_records(
    SQLHANDLE hDBC)         /**< [in] RDM ODBC Connection Handle */
{
    uint64_t      start_time;
    uint64_t      end_time;
    SQLHANDLE     hSTMT;
    SQLRETURN     sqlError;

    start_time = psp_timeNanoSecs();

    /* allocate update statement */
    sqlError = SQLAllocHandle(SQL_HANDLE_STMT, hDBC, &hSTMT);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_DBC, hDBC, "Error has occured allocating update statement");
        return sqlError;
    }

    /* Prepare and Execute the update statement */
    sqlError = SQLExecDirect(hSTMT, (SQLCHAR *) "update simple set int_col = int_col + 1", SQL_NTS);
    if (!SQL_SUCCEEDED(sqlError))
    {        
        display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error has occured preparing/executing the update statement");
        (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
        return sqlError;
    }

    (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);

    end_time = psp_timeNanoSecs();

    display_elapsed_time("Updating 50,000 records:\t\t\t", start_time, end_time);
    return sqlError;
}

/*! \brief Delete all simple records from the database
 *
 * This function will delete all simple records stored in the perf02 database.
 * The function will calculate and display the total time elapsed while 
 * deleting the records.
 *
 * @return Returns a SQLRETURN value (SQL_SUCCESS if successful)
 */
static SQLRETURN delete_records(
    SQLHANDLE hDBC)         /**< [in] RDM ODBC Connection Handle */
{
    uint64_t   start_time;
    uint64_t   end_time;
    SQLHANDLE  hSTMT;
    SQLRETURN  sqlError;

    start_time = psp_timeNanoSecs();
    
    /* allocate delete statement */
    sqlError = SQLAllocHandle(SQL_HANDLE_STMT, hDBC, &hSTMT);
    if (!SQL_SUCCEEDED(sqlError))
    {
        display_error_msg(SQL_HANDLE_DBC, hDBC, "Error has occured allocating delete statement");
        return sqlError;
    }

    /* Prepare and Execute the update statement */
    sqlError = SQLExecDirect(hSTMT, (SQLCHAR *) "DELETE FROM simple", SQL_NTS);
    if (!SQL_SUCCEEDED(sqlError))
    {        
        display_error_msg(SQL_HANDLE_STMT, hSTMT, "Error has occured preparing/executing the delete statement");
        (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
        return sqlError;
    }
    (void) SQLFreeHandle(SQL_HANDLE_STMT, hSTMT);
    end_time = psp_timeNanoSecs();

    display_elapsed_time("Deleting 50,000 records:\t\t\t", start_time, end_time);
    return sqlError;
}

/*! \brief Main function for perf02 example
 *
 * The function initializes the RDM ODBC environment and runs the create, read,
 * update, and delete operations.
 *
 * @return Returns 0 on success, and 1 if there were any errors
 */
int32_t main(void)
{
    SQLRETURN sqlError;
    SQLHANDLE hDBC;
    SQLHANDLE hENV;

    /* Display the example title */
    printf("%s", RDM_EXAMPLE_DESC("Perf02 ODBC"));

    /* Initialize the TFS, task, and open/initialize the database */
    sqlError = initialize(&hENV, &hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {
        fprintf(stderr, "Application is aborting due to initialization error\n");
        cleanup(hENV, hDBC);
        return EXIT_FAILURE;
    }

    /* Add 50,000 records to the database */
    sqlError = add_records(hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {
        fprintf(stderr, "Application is aborting due to error adding records\n");
        cleanup(hENV, hDBC);
        return EXIT_FAILURE;
    }

    /* Read all of the records in the database */
    sqlError = read_records(hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {
        fprintf(stderr, "Application is aborting due to error reading records\n");
        cleanup(hENV, hDBC);
        return EXIT_FAILURE;
    }

    /* Update all of the records in the database */
    sqlError = update_records(hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {
        fprintf(stderr, "Application is aborting due to error updating records\n");
        cleanup(hENV, hDBC);
        return EXIT_FAILURE;
    }

    /* Read all of the records in the database */
    sqlError = read_records(hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {
        fprintf(stderr, "Application is aborting due to error reading records\n");
        cleanup(hENV, hDBC);
        return EXIT_FAILURE;
    }

    /* Remove all of the records in the database */
    sqlError = delete_records(hDBC);
    if (!SQL_SUCCEEDED(sqlError))
    {
        fprintf(stderr, "Application is aborting due to error deleting records\n");
        cleanup(hENV, hDBC);
        return EXIT_FAILURE;
    }

    cleanup(hENV, hDBC);

    printf("\nPress Enter to terminate");
    getchar();

    return EXIT_SUCCESS;
}
